#include "params.h"
#include "rounding.h"
#include <stdint.h>
#include <stdio.h>
#include <assert.h>

/*************************************************
* Name:        PQCLEAN_DILITHIUM2_CLEAN_power2round
*
* Description: For finite field element a, compute a0, a1 such that
*              a mod^+ Q = a1*2^D + a0 with -2^{D-1} < a0 <= 2^{D-1}.
*              Assumes a to be standard representative.
*
* Arguments:   - int32_t a: input element
*              - int32_t *a0: pointer to output element a0
*
* Returns a1.
**************************************************/
int32_t PQCLEAN_DILITHIUM2_CLEAN_power2round(int32_t *a0, int32_t a)  {
    int32_t a1;

    a1 = (a + (1 << (D - 1)) - 1) >> D;
    *a0 = a - (a1 << D);
    return a1;
}

/*************************************************
* Name:        PQCLEAN_DILITHIUM2_CLEAN_decompose
*
* Description: For finite field element a, compute high and low bits a0, a1 such
*              that a mod^+ Q = a1*ALPHA + a0 with -ALPHA/2 < a0 <= ALPHA/2 except
*              if a1 = (Q-1)/ALPHA where we set a1 = 0 and
*              -ALPHA/2 <= a0 = a mod^+ Q - Q < 0. Assumes a to be standard
*              representative.
*
* Arguments:   - int32_t a: input element
*              - int32_t *a0: pointer to output element a0
*
* Returns a1.
**************************************************/
int32_t PQCLEAN_DILITHIUM2_CLEAN_decompose_ORG(int32_t *a0, int32_t a) {
    /* TODO:
        a % Q is skipped, as it seems a<Q always. In case this
        needs to be done, then we can use a fact that Q is a
        Generalized Marsenne Prime, so modular redc is fast
        (see work by Jerome Solina and Crandall '92 algo).
    */

    assert(a>0); assert(a<Q);

    // Use Barrett reduction to calculate r0 = r % A. The
    // code calculates:
    //     floor(a/A) = floor( (a * R) / 2^M)
    //       where,
    //          M is so that 2^M>= A^2
    //          r = floor(2^M / A)
    static const uint32_t M = 36;
    // Precomputed reciprocal r = floor((2^36) / 190464
    static const uint32_t R = 360800;
    // As per spec ALPHA, A = 2*GAMMA2 = (Q-1)/88 * 2 (Dilithium2)
    static const uint32_t A = 2*GAMMA2;

    // a0
    int32_t r;
    int32_t v,w,z;
    // Barrett reduction:
    //   a0' = a mod A = a - A*floor((a*r) / 2^M)
    r = (int32_t)((((uint64_t)a)*R) >> M);
    r = a - r*A;

    v = ((A-r)>>31) & 1;
    w = ((GAMMA2 - r)>>31) & 1;
    z = (((A + GAMMA2) -r) >> 31) & 1;
//    printf("%d %d %d\n", v,w,z);
    *a0 = r - (((!z)&(v|w))*A) - (z)*2*A;

    /*
    // REDC
    if (r>(int32_t)A) {
        r -= A;
    }

    // centrize
    if (r > (int32_t)GAMMA2) {
        *a0 = (int32_t)r - A;
    } else {
        *a0 = r;
    }
    */

    // a1
    uint64_t a2 = a - *a0;

    v = ((int32_t)a2 - Q + 1);
    //a2 = (!v);
    *a0 -= !v;
    a2 = (!!v)*a2;
    // CASE: r-r0 = q-1 => r1=0, r0 = r0-1
    #if 0
    if (a2 == (Q-1)) {
        a2 = 0;
        *a0 = *a0-1;
    }
    #endif

    // divide (r-r0)/A
    // int32_t a2 = ((uint64_t)a-*a0)/(A);
    v = ((int32_t)a2-A) >> 31;
    a2 = (!v)*(((a2*R) >> M) + 1) + v*a2;
    /*
    if ( (a2 >= (A))) {
        a2 = (a2*R) >> M;
        // a2 is divisible by ALPHA=(A) and hence
        // it will always be off by one.
        a2++;
    }
    */
    return a2;
}

int32_t PQCLEAN_DILITHIUM2_CLEAN_decompose(int32_t *a0, int32_t a) {
    /* TODO:
        a % Q is skipped, as it seems a<Q always. In case this
        needs to be done, then we can use a fact that Q is a
        Generalized Marsenne Prime, so modular redc is fast
        (see work by Jerome Solina and Crandall '92 algo).
    */

    // Use Barrett reduction to calculate r0 = r % A. The
    // code calculates:
    //     floor(a/A) = floor( (a * R) / 2^M)
    //       where,
    //          M is so that 2^M>= A^2
    //          r = floor(2^M / A)
    static const uint32_t M = 36;
    // Precomputed reciprocal r = floor((2^36) / 190464
    static const uint32_t R = 360800;
    // As per spec ALPHA, A = 2*GAMMA2 = (Q-1)/88 * 2 (Dilithium2)
    static const uint32_t A = 2*GAMMA2;

    // a0
    int32_t r;
    int32_t v,w,z;
    // Barrett reduction:
    //   a0' = a mod A = a - A*floor((a*r) / 2^M)
    r = (int32_t)((((uint64_t)a)*R) >> M);
    r = a - r*A;

    v = ((A-r)>>31) & 1;
    w = ((GAMMA2 - r)>>31) & 1;
    z = (((A + GAMMA2) -r) >> 31) & 1;
    *a0 = r - (((!z)&(v|w))*A) - (z)*2*A;

    // a1
    uint64_t a2 = a - *a0;
    v = ((int32_t)a2 - Q + 1);
    *a0 -= !v;
    a2 = (!!v)*a2;
    v = ((int32_t)a2-A) >> 31;
    a2 = (!v)*(((a2*R) >> M) + 1) + v*a2;
    return a2;
}

/*************************************************
* Name:        PQCLEAN_DILITHIUM2_CLEAN_make_hint
*
* Description: Compute hint bit indicating whether the low bits of the
*              input element overflow into the high bits.
*
* Arguments:   - int32_t a0: low bits of input element
*              - int32_t a1: high bits of input element
*
* Returns 1 if overflow.
**************************************************/
unsigned int PQCLEAN_DILITHIUM2_CLEAN_make_hint(int32_t a0, int32_t a1) {
    if (a0 > GAMMA2 || a0 < -GAMMA2 || (a0 == -GAMMA2 && a1 != 0)) {
        return 1;
    }

    return 0;
}

/*************************************************
* Name:        PQCLEAN_DILITHIUM2_CLEAN_use_hint
*
* Description: Correct high bits according to hint.
*
* Arguments:   - int32_t a: input element
*              - unsigned int hint: hint bit
*
* Returns corrected high bits.
**************************************************/
int32_t PQCLEAN_DILITHIUM2_CLEAN_use_hint(int32_t a, unsigned int hint) {
    int32_t a0, a1;

    a1 = PQCLEAN_DILITHIUM2_CLEAN_decompose(&a0, a);
    if (hint == 0) {
        return a1;
    }

    if (a0 > 0) {
        if (a1 == 43) {
            return 0;
        }
        return a1 + 1;
    }
    if (a1 == 0) {
        return 43;
    }
    return a1 - 1;
}
